package PureMessage::Policy::Delay::DelayqPmdb;
use Data::Dumper;
use PureMessage::MessageStore::Dir::Config;
use strict;
use Carp qw(croak carp);
use DBI;
use POSIX 'strftime';
my $recordCounter=0;
my $recordInsertCounter=0;	
sub new {
    my $class = shift;
    my $dir = shift;
    my %opts = @_;	
    my $verbose = delete $opts{verbose} || 0;
    my $store_class = delete $opts{class};
    my $logger = delete $opts{logger};
    my $dsn = delete $opts{dsn};
    my $location = delete $opts{location} || "localhost";
    my $store_id = delete $opts{store_id} || "pmx_quarantine";

    if (%opts) {
		if ($logger) {
			$logger->log("WARNING", "Unknown message store option '$_'") for keys %opts;
		}
		else {
			carp "Unknown option '$_'" for keys %opts;
		}
    }    
    my $max_retries = $ENV{PMX_IDX_MAX_RETRIES} || 25;
    my $self = {
		pmxdir_dir => $dir,
		pmxdir_dsn => $dsn,
		pmxdir_store_id => $store_id,
		pmxdir_verbose => $verbose,
		pmxdir_logger => $logger,
		pmxdir_sth => {},  # cached statement handles
		pmxdir_commit_period => 1,
        # for msg indexing retry mechanism
        pmxdir_max_retries => $max_retries,
        pmxdir_retry_needed => 0,
    };	
	
    bless($self, $class);
    my $conf = $class->get_config($dir);
    $self->{pmxdir_location} = $conf->value("location");
    $self->{pmxdir_location_id} = $conf->value("location_id");
	$self->{pmxdir_milter_hostname} = $conf->value("milter_hostname");
    $self->_db_connect();
    return $self;
}
sub gen_location {
    my $location_name = 'localhost'; 
    my $location_id = 0;
	my $milterhostname = 'localhost';
    eval {
        require PureMessage::Resources::Location;
        my $location = PureMessage::Resources::Location->new();
        if ($location->hostname and $location->location_name and $location->location_id) {
            $location_name = $location->location_name();
            $location_id = $location->location_id();
			$milterhostname = $location->hostname();
		}
        else {
            warn "Using defaults for location_name, location_id " .
                 "($location_name, $location_id)";
        }
    };
    warn $@ if $@;
    return ($location_name, $location_id,$milterhostname);
}
sub get_config {
    my $class = shift;
    my $dir = shift;
    my $config = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf");
    unless ($config->value("location") and $config->value("hostname") and $config->value("location_id")) {

        # get writable cfg object
        $config = PureMessage::MessageStore::Dir::Config->new(
            "$dir/db.conf", 
	    writable => 1,
            timeout => 5
        );
        my ($location_name, $location_id,$milterhostname) = gen_location();
        $config->value("location" => $location_name)
            unless $config->value("location");
        $config->value("location_id" => $location_id)
            unless $config->value("location_id");
		$config->value("milter_hostname" => $milterhostname)
            unless $config->value("milter_hostname");
        $config->save;

        # now we need a read-only config
        $config = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf");
    }
    return $config;
}
# establish new db connection
sub connect_new {
    my ($self, $connect_cached) = @_;
    my $dsn = $self->{pmxdir_dsn};
    my $dbh;
	# NOTE: We ensure that all database handles open using exactly the same
    # connection arguments so that DBI handles can be shared among multiple
    # packages.  See the following docs for more details.
    # http://perl.apache.org/docs/1.0/guide/databases.html#Opening_Connections_With_Different_Parameters

    my $connect_method = $connect_cached ? "connect_cached" : "connect";
    $dbh = DBI->$connect_method($dsn, "", "",
			{
			    RaiseError => 0,
			    PrintError => 0,
			    AutoCommit => 0,
			});
    unless ($dbh){
		$self->log("ERROR",	"Unable to connect to DB using dsn '$dsn'");
		return ;
	}
    $dbh->{RaiseError}  = 1;
    $dbh->{PrintError}  = 0; 
    $dbh->{AutoCommit}  = 0; 
    $dbh->{HandleError} = undef;
	return $dbh;
}

# keeping this global prevents this variable from going through the
# expensive initialization; this helps in CGI environment
my $pmdb_cf;

sub _db_connect {
    my $self = shift;
    my $store_id = $self->{pmxdir_store_id};
	my $dsn = $self->{pmxdir_dsn};
	if (not $dsn) {
	    require PureMessage::Database::Config;
        $pmdb_cf = PureMessage::Database::Config->new() unless $pmdb_cf;
        $dsn = $pmdb_cf->store_value($store_id, "dsn");
        $self->{pmxdir_dsn} = $dsn;
	 }
    my $connect_cached = $ENV{MOD_PERL} ? 0 : 1;
    return $self->{pmxdir_dbh} = $self->connect_new($connect_cached);
}

sub DESTROY {
    my $self = shift;
    delete $self->{pmxdir_sth};  # close all cache statement handles
    if (my $dbh = $self->{pmxdir_dbh}) {
        $dbh->rollback if (ref($dbh) eq 'Apache::DBI::db');
	$dbh->disconnect;
    }
}

# Validates config parameters required for indexing; returns false if
# indexing can not proceed with this configuration.
sub _params_valid {
    my $self = shift;
    my $dsn      = $self->{pmxdir_dsn};
    my $dir      = $self->{pmxdir_dir};
    my $location = $self->{pmxdir_location};
    my $location_id = $self->{pmxdir_location_id};

    # location must exist and be a scalar value
    if ((not $location) or (ref $location)) {
        $self->log("ERROR",
                   "Missing or invalid parameter 'location' in $dir/db.conf");
        return 0;
    }

    # location, location_id may have default values only if we're
    # co-located with the DBMS
    if ( ($dsn =~ /host=/) and
         ($dsn !~ /\Qhost=$location\E/) ) {

        if ($location eq "localhost") {
            $self->log("ERROR",
                       "Invalid parameter 'location' in $dir/db.conf: " .
                       "unable to use default value ('localhost') when dsn is set to '$dsn'");
            return 0;
        }

        if ($location_id == 0) {
            $self->log("ERROR",
                       "Invalid parameter 'location_id' in $dir/db.conf: " .
                       "unable to use default value (0) when dsn is set to '$dsn'");
            return 0;
        }
    }
    return 1;
}

sub process_delayed_files{
	my $self=shift;
	my $dir=$self->{pmxdir_dir};
	my $dbh = $self->{pmxdir_dbh};	
	unless ($dbh){
		$self->log("ERROR",	"database connection not established, can not process delayed db_entries files");
		return ;
	}
	my $delayq_folder=$dir.'/delay/db_entries/';
	unless ($delayq_folder) {
        my $msg = " delayed db_entries folder does not exist";
		$self->log("ERROR", $msg);
        croak $msg;
    }
	$self->_db_connect unless $self->{pmxdir_dbh};
    croak "Invalid configuration; see delay.log for more information"
        unless $self->_params_valid();
	$recordCounter=0;
	$recordInsertCounter=0;	
	opendir(my $d, $delayq_folder) || die "Can't read '$delayq_folder' directory: $!";
    my @files = grep !/^\./, readdir($d);
	closedir($d);
	my ($currMin)=(localtime)[1];	
	foreach my $file (@files) {
		my ($first,$second)=split('\.', $file, 2);
		if($first != ($currMin) && $first != (($currMin+1)%60) && $first != (($currMin+2)%60)){
			$self->prepare_delayq_data_from_file($file);
		}else{
			my $msg = "File:$file is skiped for indexing it will be retried.";
			$self->log("DEBUG", $msg);
		}
	}
	$self->done() if scalar(@files);
	my $msg = " No of record found in file: $recordCounter No of record inserted into db :$recordInsertCounter";
	$self->log("INFO", $msg);
}
sub prepare_delayq_data_from_file{
	my $self=shift;
	my $filename = shift || 'mail.txt';
	my $recordhash={};
	my $messageHash={};
	my $dir=$self->{pmxdir_dir};
	$filename=$dir."/delay/db_entries/".$filename;
	$self->log("DEBUG", "processing Delayed file: $filename for database entry\n");
    if (open(my $fh, '<:encoding(UTF-8)', $filename)) {
		my $columncounter=0;
		while (my $row = <$fh>) {
			chomp $row;
			if($row eq ''){
				$columncounter++;			
				next;
			}
			if($columncounter==0 ){
				$recordhash->{Global_ID}=$row;
				$columncounter++;
			}elsif($columncounter==1){
				$recordhash->{Minute_dir}=$row;
				$columncounter++;
			}elsif($columncounter==2){	
				#my $min=int($row/60000);
				$recordhash->{Delay_min}=$row;
				$columncounter++;
			}elsif($columncounter==3){
				$recordhash->{Message_size}=$row;
				$columncounter++;
			}elsif($columncounter==4){
				my $mdate = strftime '%Y-%m-%d %H:%M:%S', gmtime($row);
				$recordhash->{mdate}=$mdate;
				$columncounter++;
			}elsif($columncounter==5){
				$recordhash->{Header_From}=$row;
				$columncounter++;
			}elsif($columncounter==6){
				$recordhash->{Envelope_From}=$row;
				$columncounter++;
			}
			elsif($columncounter==7){
				my $noofline=$row;
				my $count=0;
				$recordhash->{Header_To}=[];
				while ($count++ < $noofline) {
					my $row1 = <$fh> ;
					chomp $row1;
					next if($row1 eq '');
					push(@{$recordhash->{Header_To}},$row1);
				}
				$columncounter++;
			}
			elsif($columncounter==8){
				my $noofline=$row;
				my $count=0;
				$recordhash->{Envelope_To}=[];
				while ($count++ < $noofline) {
					my $row1 = <$fh> ;
					chomp $row1;
					next if($row1 eq '');
					push (@{$recordhash->{Envelope_To}},$row1);
				}
				$columncounter++;
			}
			elsif($columncounter==9){
				my $noofline=$row;
				my $count=0;
				$recordhash->{Recipient_To}=[];
				while ($count++ < $noofline) {
					my $row1 = <$fh> ;
					chomp $row1;
					next if($row1 eq '');
					push (@{$recordhash->{Recipient_To}},$row1);
				}
				$columncounter++;
			}
			elsif($columncounter==10){
				my $noofline=$row;
				my $count=0;
				$recordhash->{C_To}=[];
				while ($count++ < $noofline) {
					my $row1 = <$fh> ;
					chomp $row1;
					next if($row1 eq '');
					push (@{$recordhash->{C_To}},$row1);
				}
				$columncounter++;
			}
			if($columncounter==11){
				$recordCounter++;
				my $row1 = <$fh>;
				$columncounter=0;
				$self->log("DEBUG", "Delayed Mail Hash:".Dumper($recordhash));
				#$messageHash->{$recordhash->{Global_ID}}=$recordhash;
				$self->_delayq_index($recordhash);
				undef($recordhash);
			}
		}
		$self->_remove_files($filename);
    } else {
		$self->log("ERROR", "Could not open file '$filename' $!");
	}
	#$self->{message}=$messageHash;
	#$self->log("DEBUG", "DB Data for Insertion".Dumper($messageHash));
}
# Wrapper for the _do_index(); encapsulates the retry and commit logic
# note that the reason for all of this complexity is the lack of support
# for savepoint in PGSQL 7.x. As soon as we migrate to 8.x this should
# be streamlined.
sub _delayq_index {
    my ($self, $message) = @_;
    eval { $self->_do_delayq_index($message) };
    if (!$@) {
        if ($self->_push_indexed($message)) {
            $self->_flush_indexed();			
        }
    }
    else {
        eval {
            my $do_die = !($self->{pmxdir_skip_message} or $self->{pmxdir_retry_needed});
            $self->_push_indexed($message) if ($self->{pmxdir_retry_needed});
            $self->_recover($message);
            $self->_flush_indexed();
            # die only if we encountered a real error
            if ($do_die) {
                my $msg= "Error while attempting to index delay queue message $message->{Global_ID}";
				$self->log("ERROR", $msg);
            }
        };
    }
    if ($@) {
        my $msg = "Unsuccessful recovery while attempting to index delay queue message $message->{Global_ID}: $@ ";
        $msg .= "Maximum count of attempts reached; giving up.";
		 my $msgdump = "Unsuccessful index delay queue message ".Dumper($message);
         $self->log("DEBUG", $msgdump);
		 $self->log("ERROR", $msg);
    }
}

# Adds a reference to the message to the list of successfully indexed
# (but not yet committed) messages; Returns true if the queue is full;
# false otherwise.
sub _push_indexed {
    my ($self, $message) = @_;
    push @{$self->{pmxdir_indexed}}, $message->{"Global_ID"};
    return ($self->{pmxdir_commit_period} <= scalar @{$self->{pmxdir_indexed}});
}

sub _flush_indexed {
    my $self = shift;
    $self->{pmxdir_dbh}->commit();
    $self->{pmxdir_indexed} = [];
}
sub _recover {
    my $self = shift;
	my $message = shift;
    my $attempt = 1;
    $self->{pmxdir_dbh}->rollback();

    my $max_retries = $self->{pmxdir_max_retries};
    while ($attempt <= $max_retries) {
        eval {$self->_do_delayq_index($message)};
        if ($@) {
            $self->{pmxdir_dbh}->rollback();
            $attempt++;
        }else {
            last;
        }
    }
    if ($attempt > $max_retries) {
        my $msg = "Unsuccessful recovery while retrying: $@";
        $self->log("ERROR", $msg);
    }
}

# called at the end of meta-indexing to commit the uncommitted insertions
# (if the count of meta-indexed messages is not evenly divisible by the
# count of messages per transaction).
sub done {
    my $self = shift;
    eval {$self->_flush_indexed();};
    if ($@) {
        $self->log("ERROR", "Error occurred while attempting the final commit: $@");
        eval {$self->{pmxdir_dbh}->rollback()};
    }
}

sub _do_delayq_index {
    my($self, $message) = @_;
    $self->_db_connect unless $self->{pmxdir_dbh};
	my $dbh = $self->{pmxdir_dbh};
    my $max_msg_cmpnnt_len = 255;
    my $m_global_id='';
    eval {
     # Get the basic message information (should all be scalars)
	$m_global_id = $message->{"Global_ID"};
	croak "Can't index a message that doesn't have a Global-ID." 
	    unless $m_global_id;
	my ($env_from_local, $env_from_dom) = 
	    split_address($message->{"Envelope_From"} || '');
	my $h_from = $message->{"Header_From"};    
	my ($h_from_local, $h_from_dom) = 
	    split_address(_clean_mail_addresses($h_from) || '');
	my $h_to = $message->{Header_To};   
	$h_to=join(",",@{$h_to});	
	my $m_milter_host = $self->{pmxdir_milter_hostname};
	my $mdate = $message->{"mdate"};
	my $message_size = $message->{Message_size};
	my $delayed_minute= $message->{"Delay_min"};
	my $delayed_dir= $message->{"Minute_dir"};
	my $location = $self->{pmxdir_location};
   # Truncate values that are too long
	for ($env_from_local, $env_from_dom, 
	    $h_from_local, $h_from_dom, $h_from, $h_to, $m_milter_host) {
	    next unless defined;
	    # make sure they fit
            if (length($_) > $max_msg_cmpnnt_len) {
                substr($_, $max_msg_cmpnnt_len) = "";
            }
	}     
	# Insert the data into the delayq_message table.
	my $insert_message = $self->{pmxdir_sth}{insert_message} ||=
	    $dbh->prepare("INSERT INTO delayq_message (m_global_id,location_name, m_date,env_from_local,
						env_from_domain,
						h_from_local,
						h_from_domain,
						h_from,
						h_to,
						m_milter_host,
						m_message_size,
						delayed_minute,
						delayed_dir						
						 ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
	$insert_message->execute($m_global_id, $location, $mdate, $env_from_local, $env_from_dom, $h_from_local, $h_from_dom, $h_from, $h_to, $m_milter_host, $message_size, $delayed_minute, $delayed_dir);
	# Get the new (database) delayq_message id
	my $get_cur_id = $self->{pmxdir_sth}{get_cur_id} ||=
	$dbh->prepare("select id from delayq_message where m_global_id = ?");
	$get_cur_id->execute($m_global_id);
	my ($cur_id) = $get_cur_id->fetchrow_array;
	$get_cur_id->finish;
	my %recipient;

	# "calculated" recipients
	foreach my $addr (@{$message->{C_To}}) {
		$addr = _clean_mail_addresses($addr);
		$recipient{$addr}{is_calculated} = 1;
	}
	# envelope recipients
    my $env_rcpt_cnt = 0;
	foreach my $addr (@{$message->{Envelope_To}}) {
            $env_rcpt_cnt++;
            $addr = _clean_mail_addresses($addr);
           $recipient{$addr}{is_envelope} = 1;
	}
	$self->{pmxdir_logger}->log("WARNING",
                                    "No envelope recipients for message " .
                                    $message->{Global_ID})
            if ($self->{pmxdir_logger} and (0 == $env_rcpt_cnt));
	
	# mapped recipients
	if (my @mrecip = @{$message->{Recipient_To}}) {
	    foreach my $r (@mrecip) {
		my $addr = _clean_mail_addresses($r);
                $recipient{$addr}{is_mapped} = 1;
	    }
	}
	$self->_store_delayq_recipient_info(\%recipient, $cur_id,$message->{Global_ID});
    $recordInsertCounter++;     
    };

    $self->{pmxdir_retry_needed} = 0;
    $self->{pmxdir_skip_message} = 0;
    if ($@) {
		my $err = $@;
		# When reindexing the quarantine it is possible that messages that
		# are already in the database will be added to the index_log. This
		# will cause the insert to fail when the global_id unique key is
		# violated. This is not normally an error (therefore we don't retry).
		if ($err =~ /violates unique constraint "unq_delayq_message_m_global_id"/) {
				my $msg = "Skipping index of delay queue message $message->{Global_ID}:"
				  . " message already exists in the database";
				$self->{pmxdir_skip_message} = 1;
			$self->log("DEBUG", $msg);
		}
			# in these cases we've hit a race condition and we want to retry
		elsif ($err =~ /deadlock detected/) {
				$self->{pmxdir_retry_needed} = 1;
				my $msg = "deadlock detected; retrying message $m_global_id";
			$self->log("DEBUG", $msg);
			}
		elsif ($err =~ /due to missing recipient/) {
				$self->{pmxdir_retry_needed} = 1;
				my $msg = "recipient dissapeared; retrying delay queue message $m_global_id";
				$self->log("DEBUG", $msg);
			}
		elsif ($err =~ /violates unique constraint "unq_delayq_recipient"/) {
				$self->{pmxdir_retry_needed} = 1;
				my $msg = "unq_delayq_recipient constraint violation; " .
						  "retrying delay queue message $m_global_id";
				$self->log("DEBUG", $msg);
			}
			# unknown error, no retry
		else {
			my $msg = "Index $m_global_id failed: $err";
			$self->log("DEBUG", $msg);
		}
	}
}

sub split_address {
    my $addr;
    $addr = shift if ref($addr = shift);
    my $max_addr_cmpnnt_len = 255;
    my ($local, $domain) = split(/\@/, $addr, 2);
    $local = "" unless defined $local;
    $domain = "" unless defined $domain;
    substr($local, $max_addr_cmpnnt_len) = ""
        if length($local) > $max_addr_cmpnnt_len;
    substr($domain, $max_addr_cmpnnt_len) = ""
        if length($domain) > $max_addr_cmpnnt_len;
    return ($local, $domain);
}

sub _store_delayq_recipient_info {
    my $self = shift;
    my $recipient_ref = shift;
    my $cur_id = shift;
    my $global_id = shift;
    my $max_addr_cmpnnt_len = 255;
    my $dbh = $self->{pmxdir_dbh};
	         
    my $ins_msg_recip_sth = $self->{pmxdir_sth}{insert_delayq_msg_recipient} ||=
        $dbh->prepare(<<"EOT");
insert into delayq_message_recipient
(recipient_id, message_id,is_mapped,  is_calculated,is_envelope)
select id, ?, ?, ?, ?
from delayq_recipient where local = ? and domain = ?
EOT
    my $ins_recip_sth = $self->{pmxdir_sth}{insert_delayq_recipient} ||=
        $dbh->prepare('insert into delayq_recipient (local, domain)
                       values (?, ?)');

    # see the above comments about race condition in
    # _store_qrntn_rule_hits(); the same thing applies here.
    while (my ($recipient, $info) = each %$recipient_ref) {
        my ($local, $domain) = split_address($recipient);
		my $recipient_inserted = 0;

        PER_RECIPIENT: while (1) {
            # the is_... columns can't be null
            $ins_msg_recip_sth->execute($cur_id,
                                        $info->{is_mapped} || 0,
										$info->{is_calculated} || 0,
										$info->{is_envelope} || 0,
                                        $local,
                                        $domain);
			last PER_RECIPIENT if $ins_msg_recip_sth->rows();
			if (not $recipient_inserted) {
				 $ins_recip_sth->execute($local, $domain);
                $recipient_inserted = 1;
				 next PER_RECIPIENT;
            }
			# We did insert the recipient, but it's not there;
            # this can happen only if another client deleted it
            # so we have a race condition. We'll let the retries
            # deal with it.
            croak "Unable to index message: $global_id " .
                  "due to missing recipient $recipient";
        }
	}
}

# Just remove the message from the DBI index
sub _trash_delayq_messages {
    my ($self, $deletedmsg) = @_;
	foreach my $msggid (@{$deletedmsg}) {
		$self->_trash_delayq_message($msggid);
	}
}
sub _trash_delayq_message {
    my($self, $gid) = @_;
	$self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
    my $msgs_deleted = 0;
	# message deletion will cascade to dependent data
    my $del_message = $self->{pmxdir_sth}{delete_delayq_message} ||=
	$dbh->prepare("DELETE FROM delayq_message WHERE m_global_id = ?");	    
    if ($gid) {
        eval {
            $del_message->execute($gid);
            $msgs_deleted = $del_message->rows();
			$dbh->commit();		
            if ($msgs_deleted) {
                $self->log("DEBUG", "Message $gid removed");
            }
            else {
                $self->log("DEBUG",
                "Delay Queue Message $gid could not be found in DB");
            }
        };
    }
    if ($@) {
	$dbh->rollback;
	}
    return $msgs_deleted;
}
#### db cleanup process is execute in each milter role. it should be execute for database role onle.
#### then the code should be move into database role with handling current timestamp of milter
sub _delay_db_cleanup {
    my($self) = @_;
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
    my $msgs_deleted = 0;
	my $currtimestamp = strftime '%Y-%m-%d %H:%M:%S', gmtime();
	#my $location=$self->{pmxdir_location};
	
	# message deletion will cascade to dependent data
    my $del_message = $self->{pmxdir_sth}{select_cleanup_delayq_message} ||=
	$dbh->prepare("delete from delayq_message where m_global_id in (select m_global_id from delayq_message where (m_date + (delayed_minute || 'minute')::interval) < ?)");	    
	eval {
		$del_message->execute($currtimestamp);
		$msgs_deleted = $del_message->rows();
		$dbh->commit();
		if ($msgs_deleted) {
			$self->log("DEBUG", "Message $msgs_deleted removed");
		}
		else {
			$self->log("DEBUG",	"No Delay Queue Message to cleanup");
		}
	};
    if ($@) {
		$dbh->rollback;
	}
    return $msgs_deleted;
}
sub get_selected_message_hash {
    my($self, $gid) = @_;
	my @gidArr=@{$gid};
    my @msgArr=();
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
	my $sql = "select m_global_id, location_name, m_milter_host,delayed_dir from delayq_message where m_global_id in (";
		$sql .= join(',', map { '?' } @gidArr);
		$sql .= ')';
	
    my $select_message = $self->{pmxdir_sth}{select_delayq_message} ||=
	#$dbh->prepare("select m_global_id, location_name, m_milter_host,delayed_dir from delayq_message where m_global_id in (?) )");	    
	$dbh->prepare($sql);
	eval {		
		$select_message->execute(@gidArr);
		my ( $m_global_id, $location_name, $m_milter_host,$delayed_dir);
		$select_message->bind_columns(\$m_global_id, \$location_name, \$m_milter_host,\$delayed_dir);
		while ($select_message->fetch()) {
			my $msghash={"gid"=>$m_global_id,"location"=>$location_name,"hostname"=>$m_milter_host,"delayDir"=>$delayed_dir};
			push(@msgArr,$msghash);
		}		
	};
    if ($@) {
		$self->log("DEBUG",	"selected_message_hash message detail not found");
	}
	return \@msgArr;
}
sub get_message_hash {
    my($self, $useraction) = @_;
	my $msghash={};
	if(!($useraction eq 'delete' || $useraction eq 'retry') ){
		$self->log("ERROR",	"mail retrival action must be retry|delete");
		return $msghash;
	}
	my $location=$self->{pmxdir_location};
	my $useractiondb=$useraction eq 'retry'?1:2;	
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
	my $select_message = $self->{pmxdir_sth}{select_delayq_message} ||=
	$dbh->prepare("select m_global_id, delayed_dir from delayq_message where location_name= ? and useraction=? ");	    
	eval {		
		$select_message->execute($location,$useractiondb);
		my ( $m_global_id, $delayed_dir);
		$select_message->bind_columns(\$m_global_id,\$delayed_dir);
		while ($select_message->fetch()) {
			$msghash->{"$m_global_id"}=$delayed_dir;
		}		
	};
    if ($@) {
		$self->log("DEBUG",	"message detail not found");
	}
	return $msghash;
}
sub update_useraction_message {
    my($self,$useraction, $gid) = @_;
	my @gidArr=@{$gid};
    my $msgs_updated=0;
	my $useractiondb=$useraction eq 'retry'?1:$useraction eq 'delete'?2:0;
	 $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};	
	my $sql = "update delayq_message set useraction=? where m_global_id in (";
		$sql .= join(',', map { '?' } @gidArr);
		$sql .= ')';
	my $update_message = $dbh->prepare($sql);	
	eval {
            $update_message->execute($useractiondb,@gidArr);
            $msgs_updated = $update_message->rows();
			$dbh->commit();		
            if ($msgs_updated) {
                $self->log("DEBUG", "$msgs_updated Messages marked as user action $useraction");
            }
            else {
                $self->log("DEBUG",
                "update_useraction_message Message @gidArr could not be found in DB. It may be Reinjected by delay runner");
            }
        };
   if ($@) {
		$self->log("ERROR",	"update_useraction_message failed to mark mail Db connecion issue");		
    }
	return $msgs_updated;
}
sub _remove_files {
    my $self = shift;
    my $count;
	while (my $file = shift) {
		 unless (unlink("$file")) {
            my $msg = "Failed to unlink file $file: $!";
	    $self->log("ERR", $msg);
            carp $msg;
	    next;
        }
        $count++;
    }
    return $count;
}
sub _clean_mail_addresses {
    my $v = shift;
    require Mail::Address;
    local $Mail::Address::VERBOSE = 0; # silence "Unmatched () ... " warnings
    my @addr = grep length, map lc($_->address), Mail::Address->parse($v);
    return @addr if wantarray;
    return $addr[0];
}

sub log {
    my $self = shift;
    if (my $logger = $self->{pmxdir_logger}) {
	return $logger->log(@_);
    }
    my $prio = shift;
    if ($prio eq "DEBUG" || $prio eq "INFO" || $prio eq "NOTICE") {
	return unless $self->{pmxdir_verbose};
    }

    my $msg = shift;
    $msg =~ s/^\s+//;
    $msg =~ s/\s+\z//;
    $msg =~ s/\s+/ /g;
    warn "LOG_$prio $msg\n";
}

1;


